import os
import tornado.httpserver
import tornado.ioloop
import tornado.wsgi
import sys
import re
from django.core.wsgi import get_wsgi_application

# options
DEFAULT_PORT = 7000
ARGUMENT_SET = {
    '-h': { 'single': True, 'given': False },
    '--port': { 'single': False, 'given': False, 'pattern': 'numberOnly' }
}

# in production.
# projectpath = [path for path in sys.path if 'site-packages' in path][0] + "/motty"

# in development.
projectpath = os.path.dirname(os.path.abspath(__file__))
sys.path.append(projectpath)

class InvalidOptionError(RuntimeError):
    def __init__(self, message):
        self.message = message

class NoValueDefinedError(RuntimeError):
    def __init__(self, arg):
        self.arg = arg

class ValueTypeError(RuntimeError):
    def __init__(self, message):
        self.message = message

def print_help_messages():
    print("""
Usage : runmotty <command>

1. runmotty                    run powerful motty server!
2. runmotty --port <number>    run motty server on <number> port.
""")

def evaluate_arguments(argv):
    """
    Basically this function converts argument list as map which consists of key and value.
    Key-Value data structure is more intuitive than list. If we want to get 'port' number. we can simply get this from calling 'arg_map['--port']'

    In addition, this function evaluates whether invalid arguments belongs to options or not.
    If the user enter alphabets next to '--port' option. the command must stop and tell users what is wrong.
    """
    is_value = False

    for i, arg in enumerate(argv):
        if is_value:
            is_value = False
            continue
        
        # if there is no key in argument set
        if arg not in ARGUMENT_SET.keys():
            raise InvalidOptionError("Unknown option error: '{0}'".format(arg))
        
        argument = ARGUMENT_SET[arg]
        if argument['single'] :
            argument['given'] = True
        else :
            # if parameter dependent option is not given with parameter.
            if len(argv)-1 < i+1:
                raise NoValueDefinedError(arg)

            # if value format is invalid.
            value = argv[i+1]

            if argument['pattern'] == 'numberOnly':
                match = re.search('^([0-9]+)$', value)
                if not match :
                    raise ValueTypeError("The '{0}' option parameter must be numbers".format(arg))

            argument['given'] = True
            argument['value'] = value
            is_value = True

def run_motty(argv):
    STATIC_ROOT = projectpath + '/app/static'
    os.environ['DJANGO_SETTINGS_MODULE'] = 'motty.settings'

    try:
        evaluate_arguments(argv)
        # help check.
        if ARGUMENT_SET['-h']['given'] :
            print_help_messages();
            return

        port = DEFAULT_PORT if not ARGUMENT_SET['--port']['given'] else ARGUMENT_SET['--port']['value']
        application = get_wsgi_application()
        container = tornado.wsgi.WSGIContainer(application)

        print("Hi, motty is now running on http://localhost:{0}/ \nYou can terminate it by pressing Ctrl + C on the keyboard.".format(port))
        print("\nThe help is displayed when you execute '$ runmotty -h'")

        tornado_app = tornado.web.Application([
            (r'/static/(.*)', tornado.web.StaticFileHandler, { 'path':STATIC_ROOT }),
            (r'.*', tornado.web.FallbackHandler, dict(fallback=container))
        ])

        http_server = tornado.httpserver.HTTPServer(tornado_app)
        http_server.listen(port)
        tornado.ioloop.IOLoop.instance().start()
    except KeyboardInterrupt:
        print("\n\nThe motty is stopped.")
    except InvalidOptionError as e:
        print(e.message, "\nAvailable options can be displayed by executing '$ runmotty -h' command")
    except NoValueDefinedError as e:
        print("'{0}' option's value should be provided.\nYou should enter command like this '$ runmotty {1} [value]'".format(e.arg, e.arg))
    except Exception as e:
        print("Unknown error occured.", e)

run_motty(sys.argv[1:]);


# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-11-17 08:58
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30)),
                ('url', models.CharField(max_length=50)),
                ('method', models.CharField(max_length=50)),
                ('contentType', models.CharField(max_length=100)),
                ('body', models.TextField()),
                ('created_at', models.DateTimeField(default=datetime.datetime(2017, 11, 17, 8, 58, 37, 690997))),
            ],
        ),
        migrations.CreateModel(
            name='Resource',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30)),
                ('url', models.CharField(max_length=50)),
            ],
        ),
        migrations.AddField(
            model_name='action',
            name='resource',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='app.Resource'),
        ),
    ]



from rest_framework.decorators import api_view

from django.shortcuts import render, redirect, reverse
from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from django.contrib import messages

from django.forms.models import model_to_dict
from django.db import connection
from django.core import serializers
from django.core.exceptions import ObjectDoesNotExist

from .utils import remove_last_slash
from .models import Action, Resource
from .serializers import ActionSerializer, ResourceSerializer

import json

# views
def main(request):
    return redirect(index)

def index(request):
    return render(request, 'app/index.html')

@api_view(['GET', 'POST'])
def save_action(request, resource_id, action_id=None):
    resource = Resource.objects.get(pk=resource_id)

    if request.method == 'GET':
        if action_id is None:
            return render(request,'app/action/form.html', { 
                        'resource': resource 
                   })
        else:
            action = Action.objects.get(pk=action_id)
            return render(request,'app/action/form.html', { 
                        'resource': resource,
                        'action': action
                   })
    else:
        body = request.data
        action = Action() if body.get('id') is None else Action.objects.get(pk=body.get('id'))
        serializer = ActionSerializer(action, data=body);

        if serializer.is_valid():
            serializer.save()
            messages.info(request, "The '{0}' action is successfully saved.".format(body.get('name')))
            return redirect('index_view')
        else:
            action_id = body.get('id')
            if action_id is None:
                return render(request, 'app/action/form.html', { 'resource': resource, 'errors': serializer.errors, 'form': body })
            else:
                return render(request, 'app/action/form.html', { 'resource': resource, 'action': action, 'errors': serializer.errors, 'form': body })

def action_view(request, id):
    action = Action.objects.get(pk=id)
    data = model_to_dict(action)
    data['resource'] = model_to_dict(Resource.objects.get(pk=data['resource']))
    data['editorType'] = get_editor_type(data['contentType'])

    return render(request, 'app/action/view.html', { 'id': id, 'action': data })

@api_view(['GET'])
def delete_action(request, id):
    try:
        action = Action.objects.get(pk=id)
        name = action.name
        action.delete()

        messages.info(request, "The '{0}' action is deleted.".format(name))
        return redirect('index_view')
    except ObjectDoesNotExist:
        messages.error(request, "No action to delete.")
        return redirect('index_view')

# api requests.
# resource api
@api_view(['GET'])
def resources(request):
    resource = Resource.objects.all()
    serializer = ResourceSerializer(resource, many=True)

    for idx, res in enumerate(serializer.data):
        for act_idx, act in enumerate(serializer.data[idx]['actions']):
            action = Action.objects.get(pk=act)
            action_serializer = ActionSerializer(action)
            serializer.data[idx]['actions'][act_idx] = action_serializer.data
            serializer.data[idx]['actions'][act_idx]['readableType'] = to_readable_type(action_serializer.data['contentType'])

    return JsonResponse(serializer.data, safe=False)

@api_view(['GET'])
def delete_resource(request, id):
    try:
        resource = Resource.objects.get(pk=id)
        resource.delete()

        messages.info(request, "The '{0}' resource is successfully deleted.".format(resource.name))
        return redirect('index_view')
    except ObjectDoesNotExist:
        messages.info(request, "No resource to delete.")
        return redirect('index_view')

@api_view(['POST'])
def save_resource(request):
    try:
        data = json.loads(request.body.decode('utf-8'))
        resource = Resource() if data.get('id') is None else Resource.objects.get(pk=data['id'])
        serializer = ResourceSerializer(resource, data=data) if data.get('id') is None else \
            ResourceSerializer(resource, data=data, partial=True)

        if serializer.is_valid() :
            serializer.save()
            return JsonResponse(serializer.data, safe=False)
        else:
            return JsonResponse(serializer.errors, status=400)
    except ObjectDoesNotExist:
        return HttpResponse("No resource to save.", status=404 )

# fake responses.
@csrf_exempt
def return_fake_request(request, endpoint):
    import re;

    method = request.method
    endpoint = '/' + endpoint
    result = re.match("^(?P<resource>\/[a-z0-9A-Z-_]+)(?P<action>\/[a-z0-9A-Z-_/]+)?", endpoint)

    resource_url = result.group('resource')
    action_url = result.group('action')

    if action_url is None:
        root_action = resource_url[1:]
        root_resource = Resource.objects.filter(url='/')[0]
        query = Action.objects.filter(resource=root_resource.id, url=root_action, method=method)
        if len(query) > 0:
            action = query[0]
            return HttpResponse(action.body, content_type=action.contentType)
    else:
        r_query = Resource.objects.filter(url=resource_url)
        if len(r_query) > 0:
            resource = r_query[0]
            query = Action.objects.filter(resource=resource.id, url=action_url, method=method)
            if len(query) > 0:
                action = query[0]
                return HttpResponse(action.body, content_type=action.contentType)

    return HttpResponse("Oops. There is no such action.", status=404)

# utils
def get_editor_type(content_type):
    json_editors = ['application/json', 'application/javascript']
    xml_editors = ['text/xml', 'application/xml']
    html_editors = ['text/html']

    if content_type in json_editors:
        return 'json';

    if content_type in xml_editors:
        return 'xml';

    if content_type in html_editors:
        return 'xml';
    
    return 'text';

def to_readable_type(content_type):
    return get_editor_type(content_type).upper();


from django.conf.urls import url
from . import views

urlpatterns = [
    url(r'^motty/$', views.index, name = "index_view"),
    url(r'^motty/resource/(?P<resource_id>[0-9]{1,})/action/create', views.save_action, name='create_action'),
    url(r'^motty/resource/(?P<resource_id>[0-9]{1,})/action/(?P<action_id>[0-9]{1,})/save', views.save_action, name='edit_action'),
    url(r'^motty/action/([0-9]{1,})/view', views.action_view, name='action_view'),
    url(r'^motty/api/action/([0-9]{1,})/delete$', views.delete_action, name = "delete_action"),

    # resource api for interactive app.
    url(r'^motty/api/resources', views.resources, name="api_resources"),
    url(r'^motty/api/resource$', views.save_resource, name="api_save_resource"),
    url(r'^motty/api/resource/([0-9]{1,})/delete', views.delete_resource, name="api_delete_resource"),

    # producing resource api
    url(r'(.+)', views.return_fake_request, name = "return_fake_request"),
    url(r'', views.main, name = "main"),
]


from django.db.models import Q
from rest_framework import serializers

import datetime

from .models import Action, Resource
from .utils import remove_last_slash

class ActionSerializer(serializers.ModelSerializer):
    class Meta:
        model = Action
        fields = ('id', 'resource', 'name', 'url', 'method', 'contentType', 'body', 'created_at')

    def validate(self, data):
        action_id = data.get('id')
        resource = data['resource']
        url = data['url']
        
        if resource.url == '/':
            if '/' in url:
                raise serializers.ValidationError("Action url must have just one slash at the beginning of it.")
            
            combined = resource.url + url
            resources = Resource.objects.filter(url=combined)
            if len(resources) > 0 :
                resource = resources[0]
                raise serializers.ValidationError("This url is same with the `%s` resource's url, URL must be unique cross the app." % (resource.name))
        else:
            if resource.url != '/' and url[0] != '/':
                raise serializers.ValidationError("Action url must start with '/', ex) '/all', '/1/product'")

        if action_id is not None:
            actions = Action.objects.filter(~Q(id=action_id), resource=resource.id, url=url)
            if len(actions) > 0:
                raise serializers.ValidationError("Action with this url already exists.")

        return data

class ResourceSerializer(serializers.ModelSerializer):
    class Meta:
        model = Resource
        fields = ('id', 'name', 'url', 'actions')
        read_only_fields = ('actions', )

    def validate_url(self, value):
        if value[0] != '/':
            raise serializers.ValidationError("URL must start with '/', ex) '/users'")

        when_slash_more_than_one = len(value.split('/')) > 2
        if when_slash_more_than_one:
            raise serializers.ValidationError("Resource url must have just one slash at the beginning of it.")

        return value;

    def get_validation_exclusions(self):
        exclusions = super(FavoriteListSerializer, self).get_validation_exclusions()
        return exclusions + ['actions']

def json_serializer(obj):
    """Default JSON serializer."""

    if isinstance(obj, datetime.datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S.%f')
    return str(obj)


# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-11-18 04:52
from __future__ import unicode_literals

import datetime
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('app', '0003_auto_20171118_0452'),
    ]

    operations = [
        migrations.AlterField(
            model_name='action',
            name='created_at',
            field=models.DateTimeField(default=datetime.datetime(2017, 11, 18, 4, 52, 18, 606440)),
        ),
    ]
